@@ -1,4 +1,4 @@
-# Copyrights 2008-2014 by [Mark Overmeer].
+# Copyrights 2008-2015 by [Mark Overmeer].
 #  For other contributors see ChangeLog.
 # See the manual pages for details on the licensing terms.
 # Pod stripped from pm file by OODoc 2.01.
@@ -7,7 +7,7 @@ use strict;
 
 package XML::Compile::Cache;
 use vars '$VERSION';
-$VERSION = '1.02';
+$VERSION = '1.04';
 
 use base 'XML::Compile::Schema';
 
@@ -30,14 +30,12 @@ sub init($)
     $self->{XCC_wopts}  = delete $args->{opts_writers} || [];
     $self->{XCC_undecl} = delete $args->{allow_undeclared} || 0;
 
-    $self->{XCC_rcode}  = {};  # compiled code refs
-    $self->{XCC_wcode}  = {};
     $self->{XCC_dropts} = {};  # declared opts
     $self->{XCC_dwopts} = {};
     $self->{XCC_uropts} = {};  # undeclared opts
     $self->{XCC_uwopts} = {};
 
-    $self->{XCC_readers} = {};
+    $self->{XCC_readers} = {}; # compiled code refs;
     $self->{XCC_writers} = {};
 
     $self->typemap($args->{typemap});
@@ -119,7 +117,6 @@ sub addPrefixes(@)
       : ref $first eq 'ARRAY' ? @$first
       : ref $first eq 'HASH'  ? %$first
       : error __x"prefixes() expects list of PAIRS, an ARRAY or a HASH";
-# warn "new prefixes: @pairs\n";
 
     my $a    = $self->{XCC_prefixes} ||= {};
     while(@pairs)
@@ -128,7 +125,7 @@ sub addPrefixes(@)
 
         if(my $def = $a->{$prefix})
         {   if($def->{uri} ne $ns)
-            {   error __x"prefix {prefix} already refers to {uri}, cannot use it for {ns}"
+            {   error __x"prefix `{prefix}' already refers to {uri}, cannot use it for {ns}"
                   , prefix => $prefix, uri => $def->{uri}, ns => $ns;
             }
         }
@@ -162,6 +159,26 @@ sub prefixFor($)
 }
 
 
+sub addNicePrefix($$)
+{   my ($self, $base, $ns) = @_;
+    if(my $def = $self->prefix($base))
+    {   return $base if $def->{uri} eq $ns;
+    }
+    else
+    {   $self->addPrefixes($base => $ns);
+        return $base;
+    }
+
+    $base .= '01' if $base !~ m/[0-9]$/;
+    while(my $def = $self->prefix($base))
+    {   return $base if $def->{uri} eq $ns;
+        $base++;
+    }
+    $self->addPrefixes($base => $ns);
+    $base;
+}
+
+
 sub learnPrefixes($)
 {   my ($self, $node) = @_;
     my $namespaces = $self->prefixes;
@@ -220,7 +237,7 @@ sub compileAll(;$$)
             {   my ($myns, $local) = unpack_type $type;
                 next if $usens eq $myns;
             }
-            $self->{XCC_rcode}{$type} ||= $self->compile(READER=>$type);
+            $self->reader($type);
         }
     }
 
@@ -230,7 +247,7 @@ sub compileAll(;$$)
             {   my ($myns, $local) = unpack_type $type;
                 next if $usens eq $myns;
             }
-            $self->{XCC_wcode}{$type} ||= $self->compile(WRITER => $type);
+            $self->writer($type);
         }
     }
 }
@@ -248,6 +265,7 @@ sub _same_params($$)
 
 sub reader($@)
 {   my ($self, $name) = (shift, shift);
+    my %args    = @_;
     my $type    = $self->findName($name);
     my $readers = $self->{XCC_readers};
 
@@ -262,7 +280,9 @@ sub reader($@)
     {   if(my $ur = $self->{XCC_uropts}{$type})
         {   # do not use cached version when options differ
             _same_params $ur, \@_
-                or return $self->compile(READER => $type, @_);
+                or return $args{is_type}
+                  ? $self->compileType(READER => $type, @_)
+                  : $self->compile(READER => $type, @_);
         }
         else
         {   $self->{XCC_uropts}{$type} = \@_;
@@ -272,13 +292,16 @@ sub reader($@)
          { error __x"type {name} is only declared as writer", name => $name }
     else { error __x"type {name} is not declared", name => $name }
 
-    $readers->{$type} ||= $self->compile(READER => $type, @_);
+    $readers->{$type} ||= $args{is_type}
+       ? $self->compileType(READER => $type, @_)
+       : $self->compile(READER => $type, @_);
 }
 
 
-sub writer($)
+sub writer($%)
 {   my ($self, $name) = (shift, shift);
-    my $type = $self->findName($name);
+    my %args    = @_;
+    my $type    = $self->findName($name);
     my $writers = $self->{XCC_writers};
 
     if(exists $self->{XCC_dwopts}{$type})
@@ -292,7 +315,9 @@ sub writer($)
     {   if(my $ur = $self->{XCC_uwopts}{$type})
         {   # do not use cached version when options differ
             _same_params $ur, \@_
-                or return $self->compile(WRITER => $type, @_)
+                or return $args{is_type}
+                  ? $self->compileType(WRITER => $type, @_)
+                  : $self->compile(WRITER => $type, @_);
         }
         else
         {   $self->{XCC_uwopts}{$type} = \@_;
@@ -305,7 +330,10 @@ sub writer($)
     {   error __x"type {name} is not declared", name => $name;
     }
 
-    $writers->{$type} ||= $self->compile(WRITER => $type, @_);
+    $writers->{$type} ||= $args{is_type}
+       ? $self->compileType(WRITER => $type, @_)
+       : $self->compile(WRITER => $type, @_);
+
 }
 
 sub template($$@)
@@ -395,6 +423,9 @@ sub mergeCompileOptions($$$)
                 push @{$x{$self->findName($t)}},  @a;
             }
         }
+        elsif($opt eq 'ignore_unused_tags')
+        {   $opts{$opt} = defined $opts{$opt} ? qr/$opts{$opt}|$val/ : $val;
+        }
         else
         {   $opts{$opt} = $val;
         }
@@ -408,11 +439,16 @@ sub _cleanup_hooks($)
 {   my ($self, $hooks) = @_;
     $hooks or return;
 
+    # translate prefixed type names into full names
     foreach my $hook (ref $hooks eq 'ARRAY' ? @$hooks : $hooks)
-    {   my $types = $hook->{type} or next;
-        $hook->{type} =
-           [ map {ref $_ eq 'Regexp' ? $_ : $self->findName($_)}
+    {   if(my $types = $hook->{type})
+        {   $hook->{type} =
+              [ map {ref $_ eq 'Regexp' ? $_ : $self->findName($_)}
                        ref $types eq 'ARRAY' ? @$types : $types ];
+        }
+        elsif(my $ext = $hook->{extends})
+        {   $hook->{extends} = $self->findName($ext);
+        }
     }
     $hooks;
 }
@@ -548,9 +584,6 @@ sub printIndex(@)
 sub _convertAnyTyped(@)
 {   my ($self, $type, $nodes, $path, $read) = @_;
 
-{no warnings;
-defined $read or warn join ';', caller(0);
-}
     my $key     = $read->keyRewrite($type);
     my $reader  = try { $self->reader($type) };
     if($@)
@@ -568,16 +601,18 @@ sub _convertAnySloppy(@)
 {   my ($self, $type, $nodes, $path, $read) = @_;
 
     my $key     = $read->keyRewrite($type);
+warn "SLOPPY $type ", map $_->toString(1), @$nodes;
     my $reader  = try { $self->reader($type) };
+warn "$@";
     if($@)
     {   # unknown type or untyped...
-        my @convert = map {XMLin $_} @$nodes;
+        my @convert = map XMLin($_), @$nodes;
         return ($key => @convert==1 ? $convert[0] : \@convert);
     }
     else
     {   trace "auto-convert known 'any' $type";
         my @nodes   = ref $nodes eq 'ARRAY' ? @$nodes : $nodes;
-        my @convert = map {$reader->($_)} @nodes;
+        my @convert = map $reader->($_), @nodes;
 
         ($key => @convert==1 ? $convert[0] : \@convert);
     }
